"""
__author__ = "Katharina Böttcher"
__copyright__ = "Copyright 2024, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from .conftest import abspath_conversion_tests

from click.testing import CliRunner
from freezegun import freeze_time

import ascmhl
from ascmhl import hasher


@freeze_time("2020-01-16 09:15:00")
def test_verify_renamed_files(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    with open("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl", "r+") as f:
        contents = f.readlines()
        for index, line in enumerate(contents):
            if "A/AA/AA1.txt" in line:
                contents.remove(line)
                new_line = line.replace("A/AA/AA1.txt", "A/AA/AA1_renamed.txt")
                contents.insert(index, new_line)
                contents.insert(index + 2, "<previousPath>A/AA/AA1.txt</previousPath>")
                break

        f.seek(0)
        f.writelines(contents)

    with open("/root/ascmhl/ascmhl_chain.xml", "r+") as f:
        contents = f.readlines()
        for index, line in enumerate(contents):
            if "0002" in line:
                hash_line = contents.pop(index + 1)
                new_line = "    <c4>{}</c4>\n".format(
                    hasher.hash_file("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl", "c4")
                )
                contents.insert(index + 1, new_line)
        f.seek(0)
        f.writelines(contents)

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    result = runner.invoke(
        ascmhl.commands.info,
        ["-sf", abspath_conversion_tests("/root/A/AA/AA1_renamed.txt"), "-v", abspath_conversion_tests("/root")],
    )

    assert not result.exception
    assert result.output.count("AA1.txt") == 3
    assert result.output.count("AA1_renamed.txt") == 3
    assert result.output.count("Generation 2") == 2


@freeze_time("2020-01-16 09:15:00")
def test_verify_renamed_file_but_also_changed_file(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    with open("/root/A/AA/AA1_renamed.txt", "a") as f:
        f.write("change file")
    with open("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl", "r+") as f:
        contents = f.readlines()
        for index, line in enumerate(contents):
            if "A/AA/AA1.txt" in line:
                contents.remove(line)
                new_line = line.replace("A/AA/AA1.txt", "A/AA/AA1_renamed.txt")
                contents.insert(index, new_line)
                contents.insert(index + 2, "<previousPath>A/AA/AA1.txt</previousPath>")
                break
        f.seek(0)
        f.writelines(contents)

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert result.exception


@freeze_time("2020-01-16 09:15:00")
def test_detect_renamed_files(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    fs.create_file("/root/B/B2.txt", contents="B2\n")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-dr"])
    assert not result.exception

    with open("/root/ascmhl/0003_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert fileContents.count("previousPath") == 2
        assert "AA1.txt" in fileContents
        assert "AA1_renamed.txt" in fileContents

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert not result.exception

    os.rename("/root/A/AA/AA1_renamed.txt", "/root/A/AA/AA1.txt")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-dr"])
    assert not result.exception


@freeze_time("2020-01-16 09:15:00")
def test_detect_renamed_files_different_hash(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    fs.create_file("/root/B/B2.txt", contents="B2\n")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-dr"])
    assert not result.exception

    with open("/root/ascmhl/0003_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert fileContents.count("previousPath") == 2
        assert "AA1.txt" in fileContents
        assert "AA1_renamed.txt" in fileContents

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert not result.exception


@freeze_time("2020-01-16 09:15:00")
def test_detect_renamed_files_different_hash(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    fs.create_file("/root/B/B2.txt", contents="B2\n")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-dr"])
    assert not result.exception

    with open("/root/ascmhl/0003_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert fileContents.count("previousPath") == 2
        assert "AA1.txt" in fileContents
        assert "AA1_renamed.txt" in fileContents

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert not result.exception


@freeze_time("2020-01-16 09:15:00")
def test_do_not_detect_renamed_files(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA/AA1.txt", "/root/A/AA/AA1_renamed.txt")
    fs.create_file("/root/B/B2.txt", contents="B2\n")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exception


@freeze_time("2020-01-16 09:15:00")
def test_detect_renamed_folders(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root")])
    assert not result.exception

    os.rename("/root/A/AA", "/root/A/AB")
    os.rename("/root/A/AB/AA1.txt", "/root/A/AB/AA2.txt")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-dr"])
    assert not result.exception
    assert "a renamed" in result.output

    with open("/root/A/AB/ascmhl/0003_AB_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert fileContents.count("previousPath") == 2

    result = runner.invoke(ascmhl.commands.verify, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert not result.exception

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-dr"])
    assert "a renamed" not in result.output
    assert not result.exception



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import pytest
from freezegun import freeze_time
from click.testing import CliRunner
from os.path import abspath
from pathlib import Path
import ascmhl.commands
import os
import time
import platform

# this file is automatically loaded by pytest we setup various shared fixtures here


def abspath_conversion_tests(path):
    return abspath(path)


def path_conversion_tests(path):
    return Path(path)


# The shutil.rmtree will not work with the fake filesystem on Python 3.12 (because it will not take symlinks),
# so here is an implementation for removing entire directories that will work with the fake fs
def remove_tree(path):
    """Recursively delete a directory tree, correctly handling symbolic links."""
    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            file_path = os.path.join(root, name)
            try:
                os.remove(file_path)  # Remove files
            except OSError as e:
                print(f"Error removing file {file_path}: {e}")

        for name in dirs:
            dir_path = os.path.join(root, name)
            try:
                # If it's a symbolic link, just remove the link itself
                if os.path.islink(dir_path):
                    os.remove(dir_path)
                else:
                    os.rmdir(dir_path)  # Remove empty directories
            except OSError as e:
                print(f"Error removing directory {dir_path}: {e}")

    # Finally remove the root directory (if not a symlink)
    if os.path.islink(path):
        os.remove(path)
    else:
        os.rmdir(path)


@pytest.fixture(scope="session", autouse=True)
def set_timezone():
    """Fakes the host timezone to UTC so we don't get different mhl files if the tests run on different time zones
    seems like freezegun can't handle timezones like we want"""
    if not os.name == "nt":
        os.environ["TZ"] = "UTZ"
        time.tzset()
    else:
        os.system('tzutil /s "UTC"')


@pytest.fixture(autouse=True)
def setup_environment(monkeypatch):
    def fake_hostname():
        return "myHost.local"

    monkeypatch.setattr(platform, "node", fake_hostname)
    # TODO: also patch ascmhl_tool_version ?


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def nested_mhl_histories(fs):
    # create mhl histories on different directly levels
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    fs.create_file("/root/A/AB/AB1.txt", contents="AB1\n")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/B/B1.txt", contents="B1\n")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/B/BA/BA1.txt", contents="BA1\n")
    fs.create_file("/root/B/BB/BB1.txt", contents="BB1\n")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B/BB"), "-h", "xxh64"])
    assert result.exit_code == 0


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def simple_mhl_history(fs):
    # create a simple mhl history with two files in one generation
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert result.exit_code == 0


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def simple_mhl_folder(fs):
    # create a simple folder structure with two files
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def post_house_file_structure(fs):
    runner = CliRunner()
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A1.txt", contents="A1")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A2.txt", contents="A2")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A3.txt", contents="A3")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt", contents="Sidecar1")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia/A"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B1.txt", contents="B1")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B2.txt", contents="B2")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B3.txt", contents="B3")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt", contents="Sidecar2")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia/B"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/CameraMedia/Report.pdf", contents="A1-3, B1-3")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/Sound/Takes/Sound.txt", contents="Sound")
    fs.create_file("/root/ShootingDay1/Sound/Sidecar.txt", contents="Sound Sidecar")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/Sound"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/Report.pdf", contents="A1-3, B1-3, Sound")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1"), "-h", "xxh64"])
    assert result.exit_code == 0

    # these are particularly relevant for ignore pattern testing
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A1_p.txt", contents="A1 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A2_p.txt", contents="A2 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A3_p.txt", contents="A3 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A001.ale", contents="A001 ALE Proxy")

    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A1_p.txt", contents="A1 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A2_p.txt", contents="A2 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A3_p.txt", contents="A3 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/B001.ale", contents="B001 ALE Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/Proxy", contents="Proxy")


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def post_house_file_structure_with_range(fs):
    runner = CliRunner()
    for i in range(1, 5):
        for j in range(1, 5):
            fs.create_file(f"/root/ShootingDay1/CameraMedia/A/A00{i}/A00{i}C00{j}.mov", contents=f"A00{i}C00{j}")

    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A1.txt", contents="A1")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A2.txt", contents="A2")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A3.txt", contents="A3")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt", contents="Sidecar1")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia/A"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B1.txt", contents="B1")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B2.txt", contents="B2")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/B3.txt", contents="B3")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt", contents="Sidecar2")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia/B"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/CameraMedia/Report.pdf", contents="A1-3, B1-3")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/Sound/Takes/Sound.txt", contents="Sound")
    fs.create_file("/root/ShootingDay1/Sound/Sidecar.txt", contents="Sound Sidecar")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/Sound"), "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/ShootingDay1/Report.pdf", contents="A1-3, B1-3, Sound")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1"), "-h", "xxh64"])
    assert result.exit_code == 0

    # these are particularly relevant for ignore pattern testing
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A1_p.txt", contents="A1 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A2_p.txt", contents="A2 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A3_p.txt", contents="A3 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/Proxy/A001/A001.ale", contents="A001 ALE Proxy")

    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A1_p.txt", contents="A1 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A2_p.txt", contents="A2 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/A3_p.txt", contents="A3 Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/B001/B001.ale", contents="B001 ALE Proxy")
    fs.create_file("/root/ShootingDay1/CameraMedia/Proxy", contents="Proxy")



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from click.testing import CliRunner
from freezegun import freeze_time
from .conftest import path_conversion_tests
from .conftest import abspath_conversion_tests

import ascmhl.commands


@freeze_time("2020-01-16 09:15:00")
def test_simple(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.flatten, [abspath_conversion_tests("/root"), abspath_conversion_tests("/out")]
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_add_one_file_same_hashformat(fs, simple_mhl_history):
    runner = CliRunner()

    # add a sidecar
    fs.create_file("/root/sidecar.txt", contents="sidecar\n")
    runner = CliRunner()
    runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64"])

    result = runner.invoke(
        ascmhl.commands.flatten, [abspath_conversion_tests("/root"), abspath_conversion_tests("/out")]
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_simple_two_hashformats(fs, simple_mhl_history):
    runner = CliRunner()

    # add a sidecar
    runner = CliRunner()
    runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "md5"])

    result = runner.invoke(
        ascmhl.commands.flatten, [abspath_conversion_tests("/root"), abspath_conversion_tests("/out")]
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_nested(fs, nested_mhl_histories):
    runner = CliRunner()

    result = runner.invoke(
        ascmhl.commands.flatten, ["-v", abspath_conversion_tests("/root"), abspath_conversion_tests("/out")]
    )
    assert result.exit_code == 0

    # check for files in root and sub histories
    assert (
        result.output == f"Flattening folder at path: {abspath_conversion_tests('/root')} ...\n"
        f"  created original hash for     Stuff.txt  xxh64: 94c399c2a9a21f9a\n"
        f"\n"
        f"Child History at {path_conversion_tests('A/AA')}:\n"
        f"  created original hash for     {path_conversion_tests('A/AA/AA1.txt')}  xxh64: ab6bec9ec04704f6\n"
        f"\n"
        f"Child History at B:\n"
        f"  created original hash for     {path_conversion_tests('B/B1.txt')}  xxh64: 51fb8fb099e92821\n"
        f"\n"
        f"Child History at {path_conversion_tests('B/BB')}:\n"
        f"  created original hash for     {path_conversion_tests('B/BB/BB1.txt')}  xxh64: 5c14eac4f4ad7501\n"
        f"Created new generation {path_conversion_tests('collection_2020-01-16/packinglist_root_2020-01-16_091500Z.mhl')}\n"
    )



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from freezegun import freeze_time
from click.testing import CliRunner
from .conftest import path_conversion_tests
from .conftest import abspath_conversion_tests

from ascmhl import utils
from ascmhl.history import MHLHistory
import ascmhl.commands

scenario_output_path = "examples/scenarios/Output"
fake_ref_path = "/ref"


@freeze_time("2020-01-16 09:15:00")
def test_create_succeed(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    # with open('/root/ascmhl/0001_root_2020-01-16_091500Z.mhl', 'r') as fin:
    #     print(fin.read())
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")


# TODO: this test should be rewritten as several different tests. Its scope is too broad. This should likely be moved to a hash specific file
@freeze_time("2020-01-16 09:15:00")
def test_create_directory_hashes(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")

    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0

    # a directory hash for the folder A was created
    hash_list = MHLHistory.load_from_path(abspath_conversion_tests("/root")).hash_lists[0]
    assert hash_list.find_media_hash_for_path("A").is_directory
    assert hash_list.find_media_hash_for_path("A").hash_entries[0].hash_string == "d3904ee76bba3d2a"
    # and the directory hash of the root folder is set in the header
    assert hash_list.process_info.root_media_hash.hash_entries[0].hash_string == "ca56d22f064fdf1b"

    # test that the directory-hash command creates the same directory hashes
    # FIXME: command doesn't exist any more, replace with tests of verify directory hashes command?
    #    result = CliRunner().invoke(ascmhl.commands.directory_hash, ["/root", "-v"])
    #    assert result.exit_code == 0
    #    assert "directory hash for: /root/A xxh64: ee2c3b94b6eecb8d" in result.output
    #    assert "root hash: xxh64: 15ef0ade91fff267" in result.output

    # add some more files and folders
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")
    os.mkdir("/root/emptyFolderB")

    os.mkdir("/root/emptyFolderC")
    os.mkdir("/root/emptyFolderC/emptyFolderCA")
    os.mkdir("/root/emptyFolderC/emptyFolderCB")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64"])
    assert result.exit_code == 0

    hash_list = MHLHistory.load_from_path(abspath_conversion_tests("/root")).hash_lists[-1]
    # due to the additional content the directory hash of folder A and the root folder changed
    assert hash_list.find_media_hash_for_path("A").hash_entries[0].hash_string == "cc195301a14023a9"
    assert hash_list.process_info.root_media_hash.hash_entries[0].hash_string == "4ccac5e6856ecf04"
    # empty folder all have the same directory hash
    assert hash_list.find_media_hash_for_path("emptyFolderA").hash_entries[0].hash_string == "ef46db3751d8e999"
    assert hash_list.find_media_hash_for_path("emptyFolderB").hash_entries[0].hash_string == "ef46db3751d8e999"
    # but since we also contain the file names in the dir hashes an empty folder that contains other empty folders
    # has a different directory structure hash
    assert (
        hash_list.find_media_hash_for_path("emptyFolderC").hash_entries[0].structure_hash_string == "949018e6a4932905"
    )
    # FIXME: emtpyFolderC isn't really empty is it - it has empty dirs inside of it.
    # how do we want to handle empty dirs in empty dirs.
    # the content hash stays the same
    assert hash_list.find_media_hash_for_path("emptyFolderC").hash_entries[0].hash_string == "cf4b060700272aa6"

    # test that the directory-hash command creates the same directory hashes
    # FIXME: command doesn't exist any more, replace with tests of verify directory hashes command?
    #    result = CliRunner().invoke(ascmhl.commands.directory_hash, ["/root"])
    #    assert result.exit_code == 0
    #    assert "  calculated root hash: xxh64: 5f4af3b3fd736415" in result.output

    # altering the content of one file
    with open("/root/A/A2.txt", "a") as file:
        file.write("!!")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64", "-h", "md5"]
    )
    assert f"ERROR: hash mismatch for        {path_conversion_tests('A/A2.txt')}" in result.output
    hash_list = MHLHistory.load_from_path(abspath_conversion_tests("/root")).hash_lists[-1]
    # an altered file leads to a different root directory hash
    assert hash_list.process_info.root_media_hash.hash_entries[1].hash_string == "28ed09733f793dfc"
    # structure hash stays the same
    assert hash_list.process_info.root_media_hash.hash_entries[1].structure_hash_string == "89e4debdb80cc068"

    # test that the directory-hash command creates the same root hash
    # FIXME: command doesn't exist any more, replace with tests of verify directory hashes command?
    #    result = CliRunner().invoke(ascmhl.commands.directory_hash, ["/root"])
    #    assert result.exit_code == 0
    #    assert "root hash: xxh64: adf18c910489663c" in result.output

    assert hash_list.find_media_hash_for_path("B").hash_entries[0].hash_string == "d6df246725efff6ceaee31f663a32cf8"
    assert hash_list.find_media_hash_for_path("B").hash_entries[1].hash_string == "aab0eba57cd1aca9"

    assert (
        hash_list.find_media_hash_for_path("B").hash_entries[0].structure_hash_string
        == "a21e164c1df944733e5e3d4e4ed64f90"
    )
    assert hash_list.find_media_hash_for_path("B").hash_entries[1].structure_hash_string == "fac2a2ceb0fa0c0b"

    assert hash_list.process_info.root_media_hash.hash_entries[1].hash_string == "28ed09733f793dfc"
    assert hash_list.process_info.root_media_hash.hash_entries[1].structure_hash_string == "89e4debdb80cc068"

    # rename one file
    os.rename("/root/B/B1.txt", "/root/B/B2.txt")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64", "-h", "c4"])
    assert f"ERROR: hash mismatch for        {path_conversion_tests('A/A2.txt')}" in result.output
    # in addition to the failing verification we also have a missing file B1/B1.txt
    assert f"missing file(s):\n  {path_conversion_tests('B/B1.txt')}" in result.output
    hash_list = MHLHistory.load_from_path(abspath_conversion_tests("/root")).hash_lists[-1]
    # the file name is part of the structure directory hash of the containing directory so it's hash changes
    assert (
        hash_list.find_media_hash_for_path("B").hash_entries[0].structure_hash_string
        == "c42qegPDBxh16Vqi4qFGh1EQv39nEbVmZ9R1LGkaVr1dEBRcD69pH3r5vdGDSwceQLEZc872kQho5Cforb95s2wjH8"
    )
    assert hash_list.find_media_hash_for_path("B").hash_entries[1].structure_hash_string == "7ae620e883160eb3"
    # .. and the content hash stays the same
    assert (
        hash_list.find_media_hash_for_path("B").hash_entries[0].hash_string
        == "c43X1ve8nmicwGit4fnhs428pTCV6ZjXQsorxPLNx3396oRuQFaq79iLR2ZsPoWN8yckFzZdkqZ21igH8K7rWAoDMa"
    )
    assert hash_list.find_media_hash_for_path("B").hash_entries[1].hash_string == "aab0eba57cd1aca9"

    # a renamed file also leads to a different root structure directory hash
    assert hash_list.process_info.root_media_hash.hash_entries[1].structure_hash_string == "0bba67923d19d36b"
    # and an unchanged content hash
    assert hash_list.process_info.root_media_hash.hash_entries[1].hash_string == "28ed09733f793dfc"

    # test that the directory-hash command creates the same root hash
    # FIXME: command doesn't exist any more, replace with tests of verify directory hashes command?


#    result = CliRunner().invoke(ascmhl.commands.directory_hash, [abspath_tests("/root")])
#    assert result.exit_code == 0
#    assert "root hash: xxh64: 01441cdf1803e2b8" in result.output


@freeze_time("2020-01-16 09:15:00")
def test_create_no_directory_hashes(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    os.mkdir("/root/emptyFolder")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-n"])
    assert result.exit_code == 0

    # a directory entry without hash was created for the folder A
    hash_list = MHLHistory.load_from_path(abspath_conversion_tests("/root")).hash_lists[0]
    assert hash_list.find_media_hash_for_path("A").is_directory
    assert len(hash_list.find_media_hash_for_path("A").hash_entries) == 0
    # and no directory hash of the root folder is set in the header
    assert hash_list.process_info.root_media_hash == None
    # the empty folder is still referenced even if not creating directory hashes
    assert hash_list.find_media_hash_for_path("emptyFolder").is_directory

    # removing an empty folder will cause creating a new generation to fail
    os.removedirs("/root/emptyFolder")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-n"])
    assert result.exit_code == 10
    assert "1 missing file(s):\n  emptyFolder" in result.output


def test_create_fail_altered_file(fs, simple_mhl_history):
    # alter a file
    with open("/root/Stuff.txt", "a") as file:
        file.write("!!")

    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root")])
    assert result.exit_code == 11
    assert "Stuff.txt" in result.output

    # since the file is still altered every other seal will fail as well since we compare to the original hash
    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root")])
    assert result.exit_code == 11
    assert "Stuff.txt" in result.output

    # when we now choose a new hash format we still fail but will add the new hash in the new format
    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "md5"])
    assert result.exit_code == 11
    assert "Stuff.txt" in result.output

    root_history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    stuff_txt_latest_media_hash = root_history.hash_lists[-1].find_media_hash_for_path("Stuff.txt")
    # the media hash for the Stuff.txt in the latest generation contains the failed xxh64 hash of the altered file
    assert stuff_txt_latest_media_hash.hash_entries[0].hash_format == "xxh64"
    assert stuff_txt_latest_media_hash.hash_entries[0].hash_string == "2346e97eb08788cc"
    assert stuff_txt_latest_media_hash.hash_entries[0].action == "failed"
    # and it contains NO new md5 hash value of the altered file
    assert len(stuff_txt_latest_media_hash.hash_entries) == 1

    # since we didn't add a new md5 hash for the failing file before creating a new generation will still fail for the altered file
    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "md5"])
    assert result.exit_code == 11
    assert "Stuff.txt" in result.output


def test_create_fail_missing_file(fs, nested_mhl_histories):
    """
    test that creating a new generation fails if there is a file missing on the file system that is referenced by one of the histories
    """

    root_history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    paths = sorted(root_history.set_of_file_paths())

    compare_paths = ["/root/B/B1.txt", "/root/B/BB/BB1.txt", "/root/Stuff.txt", "/root/A/AA/AA1.txt"]
    compare_paths = sorted([abspath_conversion_tests(x) for x in compare_paths])

    assert paths == compare_paths
    os.remove("/root/A/AA/AA1.txt")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root")])
    assert result.exit_code == 10
    assert f"1 missing file(s):\n  {path_conversion_tests('A/AA/AA1.txt')}" in result.output

    # the actual seal has been written to disk anyways we expect the history to contain
    # the new not yet referenced files (/root/B/BA/BA1.txt and /root/A/AB/AB1.txt) as well now
    root_history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    paths = sorted(root_history.set_of_file_paths())

    compare_paths = [
        "/root/B/B1.txt",
        "/root/B/BA/BA1.txt",
        "/root/B",
        "/root/A/AA",
        "/root/A/AB/AB1.txt",
        "/root/B/BA",
        "/root/A/AA/AA1.txt",
        "/root/A/AB",
        "/root/Stuff.txt",
        "/root/B/BB",
        "/root/A",
        "/root/B/BB/BB1.txt",
    ]
    compare_paths = sorted([abspath_conversion_tests(x) for x in compare_paths])

    # since we scan all generations for file paths we now get old files, missing files and new files here
    # as well as all entries for the directories
    assert paths == compare_paths

    # since the file /root/A/AA/AA1.txt is still missing all further seal attempts will still fail
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root")])
    assert result.exit_code == 10
    assert f"1 missing file(s):\n  {path_conversion_tests('A/AA/AA1.txt')}" in result.output


def test_create_nested_new_format(fs, nested_mhl_histories):
    """
    test that ensures that hasehs in a new format are also verified in child histories
    used to verify fix of bug: https://github.com/ascmitc/mhl/issues/48
    """

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "md5"])
    assert result.exit_code == 0

    # load one of the the nested histories and check the first media hash of the last generation
    nested_history = MHLHistory.load_from_path("/root/A/AA")
    media_hash = nested_history.hash_lists[-1].media_hashes[0]

    # assure that there is a hash entry for the verification of the original hash (hash formats are sorted alphabetically)

    assert media_hash.hash_entries[0].action == "verified"  # formerly 'new'
    assert media_hash.hash_entries[0].hash_format == "md5"

    assert media_hash.hash_entries[1].action == "verified"
    assert media_hash.hash_entries[1].hash_format == "xxh64"


def test_creator_info(fs, simple_mhl_history):
    # test comment
    result = CliRunner().invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "--comment", "a super comment"]
    )
    assert result.exit_code == 0
    result = CliRunner().invoke(ascmhl.commands.info, ["-v", abspath_conversion_tests("/root")])
    assert result.exit_code == 0
    assert "a super comment" in result.output

    # test location
    result = CliRunner().invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "--location", "Munich"])
    assert result.exit_code == 0
    result = CliRunner().invoke(ascmhl.commands.info, ["-v", abspath_conversion_tests("/root")])
    assert result.exit_code == 0
    assert "Munich" in result.output

    # test author
    result = CliRunner().invoke(
        ascmhl.commands.create,
        [
            abspath_conversion_tests("/root"),
            "--author_name",
            "Franz",
            "--author_email",
            "franz@example.com",
            "--author_phone",
            "123-4567",
            "--author_role",
            "Data Manager",
        ],
    )
    assert result.exit_code == 0
    result = CliRunner().invoke(ascmhl.commands.info, ["-v", abspath_conversion_tests("/root")])
    assert result.exit_code == 0
    assert "Franz" in result.output
    assert "franz@example.com" in result.output
    assert "123-4567" in result.output
    assert "Data Manager" in result.output


def test_create_mulitple_hashformats(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-n", "-h", "md5", "-h", "sha1"]
    )
    assert result.exit_code == 0

    assert f"{path_conversion_tests('A/A1.txt')}  md5: fe6975a937016c20b43b17540e6c6246" in result.output
    assert f"{path_conversion_tests('A/A1.txt')}  sha1: 4a5b95edbea7de5ed2367432645df88cd4f1d1b6" in result.output


def test_create_mulitple_hashformats_no_dash_n(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "md5", "-h", "sha1"])
    assert result.exit_code == 0

    assert f"{path_conversion_tests('A/A1.txt')}  md5: fe6975a937016c20b43b17540e6c6246" in result.output
    assert f"{path_conversion_tests('A/A1.txt')}  sha1: 4a5b95edbea7de5ed2367432645df88cd4f1d1b6" in result.output


@freeze_time("2020-01-16 09:15:00")
def test_create_mulitple_hashformats_double_hashformat(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "c4"])
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "md5", "-h", "sha1"])

    # check if mhl file exists
    mhlfilepath = "/root/ascmhl/0003_root_2020-01-16_091500Z.mhl"
    assert os.path.isfile(mhlfilepath)

    # check if mhl file validates
    result = runner.invoke(ascmhl.commands.xsd_schema_check, [abspath_conversion_tests(mhlfilepath)])
    if result.exit_code != 0:
        print(result.output)

    assert result.exit_code == 0



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import datetime
import os
import platform

import click
import pathspec
from lxml import etree

from . import logger
from . import errors
from . import ignore
from . import utils
from . import hasher
from .ignore import MHLIgnoreSpec
from .__version__ import (
    ascmhl_supported_hashformats,
    ascmhl_folder_name,
    ascmhl_tool_name,
    ascmhl_tool_version,
    ascmhl_default_hashformat,
)
from .generator import MHLGenerationCreationSession
from .hasher import hash_file, DirectoryHashContext, multiple_format_hash_file
from .hashlist import MHLMediaHash, MHLCreatorInfo, MHLProcessInfo, MHLTool, MHLProcess, MHLAuthor
from .history import MHLHistory
from .traverse import post_order_lexicographic
from typing import Dict
from collections import namedtuple
from .utils import convert_local_path_to_posix

from .utils import check_path_is_absolute_to_history


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
# general options
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
# FIXME: refactor to allow for multiple hash formats
@click.option(
    "--hash_format",
    "-h",
    type=click.Choice(ascmhl_supported_hashformats),
    multiple=True,
    default=[ascmhl_default_hashformat],
    help="Algorithm",
)
@click.option(
    "--no_directory_hashes",
    "-n",
    default=False,
    is_flag=True,
    help="Skip creation of directory hashes, only reference directories without hash",
)
@click.option(
    "--detect_renaming",
    "-dr",
    default=False,
    is_flag=True,
    help="Detect automatically renamed files",
)
# creatorinfo values
@click.option(
    "--author_name",
    default=None,
    help="Name value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_email",
    default=None,
    help="Email value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_phone",
    default=None,
    help="Phone value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_role",
    default=None,
    help="Role value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--location",
    default=None,
    help="Value for the <location> element in the <creatorinfo> element",
)
@click.option(
    "--comment",
    default=None,
    help="Value for the <comment> element in the <creatorinfo> element",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=True,
    type=click.Path(exists=True),
    help="Record single file, no completeness check (multiple occurrences possible for adding multiple files",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
def create(
    root_path,
    verbose,
    hash_format,
    no_directory_hashes,
    detect_renaming,
    single_file,
    ignore_list,
    ignore_spec_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
):
    """
    Create a new generation for a folder or file(s)

    \b
    The create command hashes all files given and creates a new generation in the
    mhl-history with records for all hashed files. The command compares the hashes
    against the hashes stored in previous generations if available.
    """
    # distinguish different behavior for entire folder vs single files
    if single_file is not None and len(single_file) > 0:
        create_for_single_files_subcommand(
            root_path,
            verbose,
            detect_renaming,
            hash_format,
            single_file,
            author_name,
            author_email,
            author_phone,
            author_role,
            location,
            comment,
            ignore_list,
            ignore_spec_file,
        )
        return
    create_for_folder_subcommand(
        root_path,
        verbose,
        detect_renaming,
        hash_format,
        no_directory_hashes,
        author_name,
        author_email,
        author_phone,
        author_role,
        location,
        comment,
        ignore_list,
        ignore_spec_file,
    )
    return


def create_for_folder_subcommand(
    root_path,
    verbose,
    detect_renaming,
    hash_formats,
    no_directory_hashes,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    # command formerly known as "seal"
    """
    Creates a new generation with all files in a folder hierarchy.

    ROOT_PATH: the root path to use for the asc mhl history

    All files are hashed and will be compared to previous records in the `asc-mhl` folder if they exists.
    The command finds files that are registered in the `asc-mhl` folder but that are missing in the file system.
    Files that are existent in the file system but are not registered in the `asc-mhl` folder yet, are registered
    as new entries in the newly created generation(s).
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"Creating new generation for folder at path: {root_path} ...")

    existing_history = MHLHistory.load_from_path(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()
    renamed_files = existing_history.renamed_path_with_previous_path()
    not_found_paths = {p if renamed_files.get(p, None) is None else renamed_files[p] for p in not_found_paths}
    new_paths = set()
    missing_asc_mhl_folder = set()

    # create the ignore specification
    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    # start a verification session on the existing history
    session = MHLGenerationCreationSession(existing_history, ignore_spec)

    # update the ignore spec and include ignores from nested histories
    ignore_spec = get_ignore_spec_including_nested_ignores(existing_history, ignore_list, ignore_spec_file)

    num_failed_verifications = 0
    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    # the mapping lookups will follow the dictionary format of [string: [hash_format: hash_value]] where string
    # is a file sub-path
    dir_content_hash_mapping_lookup = {}
    dir_structure_hash_mapping_lookup = {}
    hash_format_list = sorted(hash_formats)

    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        # generate directory hashes
        dir_hash_context_lookup = {}

        if not no_directory_hashes:
            # Create a DirectoryHashContext for each hash format and store in the lookup
            for hash_format in hash_format_list:
                dir_hash_context_lookup[hash_format] = DirectoryHashContext(hash_format)
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            if file_path in not_found_paths:
                not_found_paths.discard(file_path)
            else:
                new_paths.add(file_path)

            if is_dir:
                if not no_directory_hashes:
                    path_content_hash_lookup = dir_content_hash_mapping_lookup.pop(file_path)
                    path_structure_hash_lookup = dir_structure_hash_mapping_lookup.pop(file_path)

                    for hash_format, dir_hash_context in dir_hash_context_lookup.items():
                        dir_hash_context.append_directory_hashes(
                            file_path,
                            path_content_hash_lookup[hash_format],
                            path_structure_hash_lookup[hash_format],
                        )
            else:
                seal_result = seal_file_path(existing_history, file_path, hash_format_list, session)

                for hash_format, result_tuple in seal_result.items():
                    dir_hash_context = None

                    if not no_directory_hashes:
                        dir_hash_context = dir_hash_context_lookup[hash_format]

                    hash_string = result_tuple.hash_value
                    success = result_tuple.success
                    if not success:
                        num_failed_verifications += 1
                    if dir_hash_context is not None:
                        dir_hash_context.append_file_hash(file_path, hash_string)

        # Calculate the directory hashes for each format
        dir_content_hash_lookup = {}
        dir_structure_hash_lookup = {}

        if not no_directory_hashes:
            for hash_format, dir_hash_context in dir_hash_context_lookup.items():
                dir_content_hash = dir_hash_context.final_content_hash_str()
                dir_structure_hash = dir_hash_context.final_structure_hash_str()

                if dir_content_hash_mapping_lookup and folder_path in dir_content_hash_mapping_lookup.keys():
                    dir_content_hash_mapping_lookup[folder_path][hash_format] = dir_content_hash
                else:
                    dir_content_hash_mapping_lookup[folder_path] = {hash_format: dir_content_hash}

                if dir_structure_hash_mapping_lookup and folder_path in dir_structure_hash_mapping_lookup.keys():
                    dir_structure_hash_mapping_lookup[folder_path][hash_format] = dir_structure_hash
                else:
                    dir_structure_hash_mapping_lookup[folder_path] = {hash_format: dir_structure_hash}

                dir_content_hash_lookup[hash_format] = dir_content_hash
                dir_structure_hash_lookup[hash_format] = dir_structure_hash

        modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(folder_path))

        session.append_multiple_format_directory_hashes(
            folder_path, modification_date, dir_content_hash_lookup, dir_structure_hash_lookup
        )

    if len(existing_history.hash_lists) > 0:
        for ref in existing_history.hash_lists[-1].hash_list_references:
            referenced_asc_folder = os.path.join(
                os.path.dirname(existing_history.asc_mhl_path), os.path.dirname(ref.path)
            )
            # as we also ignore the ascmhl folder itself check the ignore patterns for the enclosing folder
            if not os.path.exists(referenced_asc_folder) and not ignore_spec.get_path_spec().match_file(
                os.path.dirname(referenced_asc_folder)
            ):
                missing_asc_mhl_folder.add(os.path.dirname(referenced_asc_folder))

    if detect_renaming:
        found_file_paths = set()
        for new_path in new_paths:
            for not_found_path in not_found_paths:
                # find hashes to not_found_path and new_path
                not_found_path_history, relative_not_found_path = existing_history.find_history_for_path(
                    existing_history.get_relative_file_path(not_found_path)
                )
                not_found_path_hash = not_found_path_history.find_first_hash_entry_for_path(relative_not_found_path)

                new_path_history, new_path_media_hash = None, None
                for history, hash_list in session.new_hash_lists.items():
                    new_path_media_hash = hash_list.find_media_hash_for_path(history.get_relative_file_path(new_path))
                    if new_path_media_hash is not None:
                        new_path_history = history
                        break
                new_path_hash = new_path_media_hash.find_hash_entry_for_format(not_found_path_hash.hash_format)
                # compare found hashes
                if new_path_hash:
                    if new_path_hash.hash_string == not_found_path_hash.hash_string:
                        if os.path.basename(new_path) != os.path.basename(not_found_path):
                            logger.info(
                                "a renamed {} was detected: from {} to {}".format(
                                    "folder" if os.path.isdir(new_path) else "file",
                                    relative_not_found_path,
                                    existing_history.get_relative_file_path(new_path),
                                )
                            )
                        if new_path_media_hash.path == ".":
                            root_hash = session.new_hash_lists[
                                new_path_history.parent_history
                            ].find_media_hash_for_path(new_path_history.parent_history.get_relative_file_path(new_path))
                            if root_hash:
                                root_hash.previous_path = relative_not_found_path
                        else:
                            new_path_media_hash.previous_path = relative_not_found_path
                        if not_found_path in missing_asc_mhl_folder:
                            if os.path.exists(os.path.join(new_path, ascmhl_folder_name)):
                                missing_asc_mhl_folder.discard(not_found_path)
                            else:
                                missing_asc_mhl_folder.discard(not_found_path)
                                missing_asc_mhl_folder.add(new_path)
                        found_file_paths.add(not_found_path)
                else:
                    old_hash_format_for_new_path = hasher.hash_file(
                        os.path.join(root_path, new_path), not_found_path_hash.hash_format
                    )
                    if old_hash_format_for_new_path == not_found_path_hash.hash_string:
                        if os.path.basename(new_path) != os.path.basename(not_found_path):
                            logger.info(
                                "a renamed {} was detected: from {} to {}".format(
                                    "folder" if os.path.isdir(new_path) else "file",
                                    relative_not_found_path,
                                    existing_history.get_relative_file_path(new_path),
                                )
                            )
                        new_path_media_hash.previous_path = relative_not_found_path
                        found_file_paths.add(not_found_path)
        not_found_paths = not_found_paths - found_file_paths
    commit_session(session, author_name, author_email, author_phone, author_role, location, comment)

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()

    if exception:
        raise exception

    if len(missing_asc_mhl_folder) > 0:
        raise errors.NoMHLHistoryException(", ".join(missing_asc_mhl_folder))


def create_for_single_files_subcommand(
    root_path,
    verbose,
    detect_renaming,
    hash_formats,
    single_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    # command formerly known as "record"
    """
    Creates a new generation with the given file(s) or folder(s).

    \b
    ROOT_PATH: the root path to use for the asc mhl history
    single_file: one or multiple paths to files or folders to be recorded

    All files that are specified or the files inside a specified folder are hashed and will be compared
    to previous records in the `asc-mhl` folder if they are recorded in the history already.

    The following files will not be handled by this command:

    \b
    * files that are referenced in the existing ascmhl history but not specified as input
    * files that are neither referenced in the history nor specified as input
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    assert len(single_file) != 0

    existing_history = MHLHistory.load_from_path(root_path)
    # start a creation session on the existing history
    session = MHLGenerationCreationSession(existing_history)

    num_failed_verifications = 0

    hash_format_list = sorted(hash_formats)

    for path in single_file:
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)
        if os.path.isdir(path):
            for folder_path, children in post_order_lexicographic(path, session.ignore_spec.get_path_spec()):
                for item_name, is_dir in children:
                    file_path = os.path.join(folder_path, item_name)
                    if is_dir:
                        continue
                    seal_result = seal_file_path(existing_history, file_path, hash_format_list, session)
                    # Determine success based on the first format in the list
                    # TODO: Consider checking all results.  Would it be practical to do so?
                    # For instance, are we concerned about one format failing while another one succeeds?
                    success = seal_result[hash_format_list[0]].success
                    if not success:
                        num_failed_verifications += 1
        else:
            seal_result = seal_file_path(existing_history, path, hash_format_list, session)
            success = seal_result[hash_format_list[0]].success
            if not success:
                num_failed_verifications += 1

    commit_session(session, author_name, author_email, author_phone, author_role, location, comment)

    if num_failed_verifications > 0:
        raise errors.VerificationFailedException()


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
# subcommand
@click.option(
    "--directory_hash",
    "-dh",
    default=False,
    is_flag=True,
    help="Just create and verify directory hashes.",
)
# options for directory_hash
@click.option(
    "--calculate_only",
    "-co",
    default=False,
    is_flag=True,
    help="Only calculate and print hashes, don't compare against history.",
)
@click.option(
    "--root_only",
    "-ro",
    default=False,
    is_flag=True,
    help=(
        "Only calculate and print root directory hash, don't compare against history (only applies to"
        " --directory_hash)."
    ),
)
# options
@click.option(
    # FIXME: Update to permit multiple hash formats
    "--hash_format",
    "-h",
    type=click.Choice(ascmhl_supported_hashformats),
    multiple=False,
    help="Algorithm for directory hashes",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=False,
    type=click.Path(),
    help="Verify a single file",
)
@click.option(
    "--packing_list", "-pl", default=None, type=click.Path(exists=True), help="Verify against an external packing list"
)
def verify(
    root_path,
    verbose,
    directory_hash,
    hash_format,
    single_file,
    packing_list,
    ignore_list,
    ignore_spec_file,
    calculate_only,
    root_only,
):
    """
    Verify a folder, single file(s), or a directory hash

    \b
    The verify command is used to check files in the file system with records in
    the ASC MHL history. All given files are hashed and hash values are compared
    against hash values stored in the ASC MHL history. Missing files or additional
    files in the file system are reported as errors. No new ASC MHL file /
    generation is created.
    """

    if packing_list is not None:
        verify_entire_folder(
            root_path, verbose, single_file, packing_list, ignore_list, ignore_spec_file, calculate_only
        )
        return

    if directory_hash is True:
        verify_directory_hash_subcommand(
            root_path, verbose, hash_format, ignore_list, ignore_spec_file, calculate_only, root_only
        )
        return

    verify_entire_folder(root_path, verbose, single_file, None, ignore_list, ignore_spec_file)
    return


def verify_entire_folder(
    root_path, verbose, single_file, packing_list_path, ignore_list=None, ignore_spec_file=None, calculate_only=None
):
    """
    Checks MHL hashes from all generations / a packing list against all file hashes.

    ROOT_PATH: the root path to use for the asc mhl history
    packing_list (opt): the path to the packing list

    Traverses through the content of a folder, hashes all found files and compares ("verifies") the hashes
    against the records in the asc-mhl folder/packing list. The command finds all files that are existent in
    the file system but are not registered in the `asc-mhl` folder/packing list yet, and all files that are
    registered in the `asc-mhl` folder/packing list but that are missing in the file system.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    if single_file is not None and not os.path.isabs(single_file):
        single_file = os.path.join(root_path, single_file)

    logger.verbose(f"check folder at path: {root_path}")

    if packing_list_path is not None:
        existing_history = MHLHistory.load_from_packing_list_path(packing_list_path, root_path)
    else:
        existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()
    renamed_files = existing_history.renamed_path_with_previous_path()
    not_found_paths = {p if renamed_files.get(p, None) is None else renamed_files[p] for p in not_found_paths}

    num_failed_verifications = 0
    num_new_files = 0

    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    found_single_file = False

    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            not_found_paths.discard(file_path)
            relative_path = existing_history.get_relative_file_path(file_path)
            history, history_relative_path = existing_history.find_history_for_path(relative_path)
            if is_dir:
                # TODO: find new directories here
                continue

            for hash_list in existing_history.hash_lists:
                for media_hash in hash_list.media_hashes:
                    if media_hash.path != history_relative_path:
                        continue
                    history_relative_path = media_hash.previous_path or history_relative_path
                    break

            if single_file is None or os.path.realpath(single_file) == os.path.realpath(file_path):
                # check if there is an existing hash in the other generations and verify
                original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

                # in case there is no original hash entry continue
                if original_hash_entry is None:
                    logger.error(f"found new file {relative_path}")
                    num_new_files += 1
                    continue

                # create a new hash and compare it against the original hash entry
                current_hash = hash_file(file_path, original_hash_entry.hash_format)
                if original_hash_entry.hash_string == current_hash:
                    logger.verbose(f"verification ({original_hash_entry.hash_format}) of file {relative_path}: OK")
                else:
                    logger.error(
                        f"ERROR: hash mismatch        for {relative_path} "
                        f"old {original_hash_entry.hash_format}: {original_hash_entry.hash_string}, "
                        f"new {original_hash_entry.hash_format}: {current_hash}"
                    )
                    num_failed_verifications += 1

                found_single_file = True

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)

    if not found_single_file:
        exception = errors.SingleFileNotFoundException()

    if num_new_files > 0:
        exception = errors.NewFilesFoundException()
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()

    if exception:
        raise exception


def verify_directory_hash_subcommand(
    root_path, verbose, hash_format, ignore_list=None, ignore_spec_file=None, calculate_only=False, root_only=False
):
    """
    Checks MHL directory hashes from all generations against computed directory hashes.

    ROOT_PATH: the root path to use for the asc mhl history

    Traverses through the content of a folder, hashes all found files, create directory hashes, and compares
    ("verifies") the hashes against the directory hash records in the asc-mhl folder.
    Content directory hashes and structure directory hashes are compared individually.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"check folder at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    ignore_spec = get_ignore_spec_including_nested_ignores(existing_history, ignore_list, ignore_spec_file)

    # FIXME: Update once argument signature has been modified to supply a list of formats
    hash_formats = []

    # choose the hash format of the latest root directory hash
    if hash_format is None:
        generation = -1
        # inspect the history and use all documented algorithms as the basis of verification
        for hash_list in existing_history.hash_lists:
            if hash_list.generation_number > generation:
                # add each hash entry's format to the list of formats
                if len(hash_list.process_info.root_media_hash.hash_entries) > 0:
                    for entry in hash_list.process_info.root_media_hash.hash_entries:
                        entry_hash_format = entry.hash_format
                        # do not permit duplicate entries in the list
                        if entry_hash_format not in hash_formats:
                            hash_formats.append(entry_hash_format)
        if not hash_formats:
            hash_formats.append("c4")
            logger.verbose(f"default hash format: c4")
        else:
            logger.verbose(f"hash format from latest generation with directory hashes: {hash_format}")
    else:
        hash_formats.append(hash_format)
        logger.verbose(f"hash format: {hash_format}")

    # start a verification session on the existing history
    hash_format_list = sorted(hash_formats)
    # a lookup containing the number of failures detected per hash format - will match the format Dict[str, int]
    failures_per_format_lookup = {}

    # an inner function responsible for adding format failures to the lookup
    def add_detected_failure_for_format(failed_hash_format):
        nonlocal failures_per_format_lookup
        if failures_per_format_lookup:
            if failed_hash_format in failures_per_format_lookup.keys():
                failures_per_format_lookup[failed_hash_format] = failures_per_format_lookup[failed_hash_format] + 1
            else:
                failures_per_format_lookup[failed_hash_format] = 1
        else:
            failures_per_format_lookup = {failed_hash_format: 1}

    session = MHLGenerationCreationSession(existing_history)

    num_failed_verifications = 0
    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    dir_content_hash_mappings = {}
    dir_structure_hash_mappings = {}
    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        # generate directory hashes - will match the format dict[str, DirectoryHashContext]
        dir_hash_context_lookup = {}

        # create a DirectoryHashContext for each hash format and store in the lookup
        for hash_format in hash_format_list:
            dir_hash_context_lookup[hash_format] = DirectoryHashContext(hash_format)

        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            if is_dir:
                relative_path = existing_history.get_relative_file_path(file_path)
                history, history_relative_path = existing_history.find_history_for_path(relative_path)
                # check if there are directory hashes in the generations
                directory_hash_entries = history.find_directory_hash_entries_for_path(history_relative_path)

                content_hash_lookup = dir_content_hash_mappings.pop(file_path)
                structure_hash_lookup = dir_structure_hash_mappings.pop(file_path)

                # Add the content and hash values to the appropriate context
                if dir_hash_context_lookup:
                    for hash_format, dir_hash_context in dir_hash_context_lookup.items():
                        content_hash = None
                        structure_hash = None
                        if content_hash_lookup:
                            content_hash = content_hash_lookup[hash_format]
                        if structure_hash_lookup:
                            structure_hash = structure_hash_lookup[hash_format]

                        dir_hash_context.append_directory_hashes(file_path, content_hash, structure_hash)

                num_successful_verifications = 0
                for directory_hash_entry in directory_hash_entries:
                    content_hash = None
                    structure_hash = None

                    if content_hash_lookup:
                        content_hash = content_hash_lookup[directory_hash_entry.hash_format]
                    if structure_hash_lookup:
                        structure_hash = structure_hash_lookup[directory_hash_entry.hash_format]

                    found_hash_format = False

                    if content_hash:
                        found_hash_format = True

                        num_current_successful_verifications = -1
                        if not root_only:
                            num_current_successful_verifications = _compare_and_log_directory_hashes(
                                relative_path, directory_hash_entry, content_hash, structure_hash
                            )

                        if num_current_successful_verifications == 2:
                            num_successful_verifications += 1
                        if num_current_successful_verifications == 1:
                            num_failed_verifications += 1
                            add_detected_failure_for_format(directory_hash_entry.hash_format)

                    if not calculate_only:
                        if not found_hash_format:
                            logger.error(
                                f"ERROR: verification of folder {relative_path}: No directory hash of type"
                                f" {hash_format} found"
                            )
                            num_failed_verifications += 1
                            add_detected_failure_for_format(directory_hash_entry.hash_format)
            else:
                # generate the hashes for the file
                file_hash_lookup = multiple_format_hash_file(file_path, hash_format_list)
                # add each hash to the appropriate context
                for hash_format, hash_value in file_hash_lookup.items():
                    dir_hash_context_lookup[hash_format].append_file_hash(file_path, hash_value)

        # all children have been handled.  create the directory hashes
        dir_content_hash_lookup = {}
        dir_structure_hash_lookup = {}

        if dir_hash_context_lookup:
            for hash_format, dir_hash_context in dir_hash_context_lookup.items():
                dir_content_hash = dir_hash_context.final_content_hash_str()
                dir_structure_hash = dir_hash_context.final_structure_hash_str()
                # add the hashes to the appropriate hash lookups to be added to the session
                dir_content_hash_lookup[hash_format] = dir_content_hash
                dir_structure_hash_lookup[hash_format] = dir_structure_hash
            # add the hash lookups to the appropriate mappings for the folder
            if dir_content_hash_lookup:
                dir_content_hash_mappings[folder_path] = dir_content_hash_lookup
            if dir_structure_hash_lookup:
                dir_structure_hash_mappings[folder_path] = dir_structure_hash_lookup

        modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(folder_path))

        logger.verbose_logging = calculate_only
        relative_path = session.root_history.get_relative_file_path(folder_path)
        if root_only and relative_path != ".":
            logger.verbose_logging = verbose

        session.append_multiple_format_directory_hashes(
            folder_path, modification_date, dir_content_hash_lookup, dir_structure_hash_lookup
        )
        logger.verbose_logging = verbose

        # compare root hashes, works differently
        if folder_path == root_path:
            for hash_list in existing_history.hash_lists:
                root_hash_entries = hash_list.process_info.root_media_hash.hash_entries
                if len(root_hash_entries) > 0:
                    for root_hash_entry in root_hash_entries:
                        hash_format = root_hash_entry.hash_format
                        found_hash_format = False
                        dir_content_hash = None
                        dir_structure_hash = None

                        if dir_content_hash_lookup:
                            dir_content_hash = dir_content_hash_lookup[hash_format]
                        if dir_structure_hash_lookup:
                            dir_structure_hash = dir_structure_hash_lookup[hash_format]

                        if dir_content_hash:
                            found_hash_format = True
                            _compare_and_log_directory_hashes(
                                ".", root_hash_entry, dir_content_hash, dir_structure_hash
                            )

                        if not calculate_only:
                            if not found_hash_format:
                                logger.error(
                                    f"ERROR: verification of root folder: No directory hash of type {hash_format} found"
                                )
    exception = None

    # check the failure lookup.  if even one format verified, consider the entire process verified
    if failures_per_format_lookup:
        if len(failures_per_format_lookup.keys()) == len(hash_format_list):
            exception = errors.VerificationDirectoriesFailedException()

    if exception:
        raise exception


def _compare_and_log_directory_hashes(
    relative_path, directory_hash_entry, calculated_content_hash_string, calculated_structure_hash_string
):
    num_successful_verifications = 0
    root_string = ""
    if hasattr(directory_hash_entry, "temp_is_root_folder") and directory_hash_entry.temp_is_root_folder:
        root_string = " (root folder in child history)"
    if (
        directory_hash_entry.hash_string == calculated_content_hash_string
        and directory_hash_entry.structure_hash_string == calculated_structure_hash_string
    ):
        if relative_path == ".":
            logger.verbose(
                f"  verification of root folder   OK (generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  verification of folder        {relative_path}{root_string} OK "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )

        num_successful_verifications += 2
    else:
        if directory_hash_entry.hash_string != calculated_content_hash_string:
            logger.error(
                f"ERROR: content hash mismatch   for {relative_path}{root_string} "
                f"old {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string}, "
                f"new {directory_hash_entry.hash_format}: {calculated_content_hash_string} "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  content hash matches for      {relative_path}{root_string} "
                f" {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string}"
                f" (generation {directory_hash_entry.temp_generation_number:04d})"
            )

        if directory_hash_entry.structure_hash_string != calculated_structure_hash_string:
            logger.error(
                f"ERROR: structure hash mismatch for {relative_path}{root_string} "
                f"old {directory_hash_entry.hash_format}: {directory_hash_entry.structure_hash_string}, "
                f"new {directory_hash_entry.hash_format}: {calculated_structure_hash_string} "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  structure hash matches for    {relative_path}{root_string} "
                f" {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string} "
                f" (generation {directory_hash_entry.temp_generation_number:04d})"
            )

        num_successful_verifications += 1

    return num_successful_verifications


# TODO def verify_single_file_subcommand(root_path, verbose):


@click.command()
@click.argument("file_path", type=click.Path(exists=True))
@click.option(
    # FIXME: Update to permit multiple hash formats
    "--hash_format",
    "-h",
    type=click.Choice(ascmhl_supported_hashformats),
    multiple=False,
    required=True,
    help="Algorithm",
)
def hash(file_path, hash_format):
    """
    Create and print a hash value for a file
    """
    result = hash_file(file_path, hash_format)
    logger.info(hash_format + " (" + file_path + ") = " + result)


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
def diff(root_path, verbose, ignore_list, ignore_spec_file):
    """
    Diff an entire folder structure

    \b
    The diff command is used to quickly compare files in the file system with
    records in the ASC MHL history. In comparison to the verify command, no
    hash values are created and compared. Missing files or additional files
    in the file system are reported as errors. No new ASC MHL file / generation
    is created.
    """
    diff_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list, ignore_spec_file)
    return


def diff_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list=None, ignore_spec_file=None):
    """
    Checks MHL hashes from all generations against all file hash entries.

    ROOT_PATH: the root path to use for the asc mhl history

    Traverses through the content of a folder. The command finds all files that are existent in the file system
    but are not registered in the `asc-mhl` folder yet, and all files that are registered in the `asc-mhl` folder
    but that are missing in the file system.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"check folder at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()
    renamed_files = existing_history.renamed_path_with_previous_path()
    not_found_paths = {p if renamed_files.get(p, None) is None else renamed_files[p] for p in not_found_paths}

    num_failed_verifications = 0
    num_new_files = 0

    ignore_spec = get_ignore_spec_including_nested_ignores(existing_history, ignore_list, ignore_spec_file)

    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            not_found_paths.discard(file_path)
            relative_path = existing_history.get_relative_file_path(file_path)
            history, history_relative_path = existing_history.find_history_for_path(relative_path)
            if is_dir:
                # TODO: find new directories here
                continue

            for hash_list in existing_history.hash_lists:
                for media_hash in hash_list.media_hashes:
                    if media_hash.path != history_relative_path:
                        continue
                    history_relative_path = media_hash.previous_path or history_relative_path
                    break

            # check if there is an existing hash in the other generations and verify
            original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

            # in case there is no original hash entry continue
            if original_hash_entry is None:
                logger.error(f"found new file {relative_path}")
                num_new_files += 1
                continue

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()
    if not exception and num_new_files > 0:
        exception = errors.NewFilesFoundException()

    if exception:
        raise exception


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.argument("destination_path", type=click.Path())
# general options
@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
@click.option(
    "--no_directory_hashes",
    "-n",
    default=False,
    is_flag=True,
    help="Skip creation of directory hashes, only reference directories without hash",
)
@click.option("ignore_list", "--ignore", "-i", multiple=True, help="A single file pattern to ignore.")
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
# creatorinfo values
@click.option(
    "--author_name",
    default=None,
    help="Name value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_email",
    default=None,
    help="Email value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_phone",
    default=None,
    help="Phone value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_role",
    default=None,
    help="Role value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--location",
    default=None,
    help="Value for the <location> element in the <creatorinfo> element",
)
@click.option(
    "--comment",
    default=None,
    help="Value for the <comment> element in the <creatorinfo> element",
)
def flatten(
    root_path,
    destination_path,
    verbose,
    no_directory_hashes,
    ignore_list,
    ignore_spec_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
):
    """
    Flatten an MHL history into one external manifest

    \b
    The flatten command iterates through the mhl-history, collects all known files and
    their hashes in multiple hash formats and writes them to a new mhl file outside of the
    iterated history.
    """
    flatten_history(
        root_path,
        destination_path,
        verbose,
        no_directory_hashes,
        author_name,
        author_email,
        author_phone,
        author_role,
        location,
        comment,
        ignore_list,
        ignore_spec_file,
    )
    return


def flatten_history(
    root_path,
    destination_path,
    verbose,
    no_directory_hashes,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"Flattening folder at path: {root_path} ...")

    existing_history = MHLHistory.load_from_path(root_path)

    # create the ignore specification
    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    # start a verification session on the existing history
    collection_history = MHLHistory.create_collection_at_path(destination_path)
    session = MHLGenerationCreationSession(collection_history, ignore_spec)

    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    dir_hash_mappings = {}

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    flatten_child_histories(existing_history, session, root_path)

    commit_session_for_collection(
        session, root_path, author_name, author_email, author_phone, author_role, location, comment
    )


def flatten_child_histories(history, session, roothistorypath, pathprefix=""):
    for hash_list in history.hash_lists:
        for media_hash in hash_list.media_hashes:
            if not media_hash.is_directory:
                for hash_entry in media_hash.hash_entries:
                    if hash_entry.action != "failed":
                        # add prefix to media path if subhistory
                        media_path = media_hash.path
                        if pathprefix != "":
                            media_path = convert_local_path_to_posix(pathprefix) + "/" + media_hash.path
                        # check if this entry is newer than the one already in there to avoid duplicate entries
                        found_media_hash = session.new_hash_lists[session.root_history].find_media_hash_for_path(
                            media_path
                        )
                        if found_media_hash == None:
                            session.append_file_hash(
                                media_path,
                                media_hash.file_size,
                                media_hash.last_modification_date,
                                hash_entry.hash_format,
                                hash_entry.hash_string,
                                action=hash_entry.action,
                                hash_date=hash_entry.hash_date,
                            )
                        else:
                            hashformat_is_already_there = False
                            for found_hash_entry in found_media_hash.hash_entries:
                                if found_hash_entry.hash_format == hash_entry.hash_format:
                                    hashformat_is_already_there = True
                            if not hashformat_is_already_there:
                                # assuming that hash_entry of same type also has same hash_value ..
                                session.append_file_hash(
                                    media_path,
                                    media_hash.file_size,
                                    media_hash.last_modification_date,
                                    hash_entry.hash_format,
                                    hash_entry.hash_string,
                                    action=hash_entry.action,
                                    hash_date=hash_entry.hash_date,
                                )

    for child_history in history.child_histories:
        childpath = child_history.get_root_path()
        childrelativepath = os.path.relpath(childpath, roothistorypath)

        logger.info(f"\nChild History at {childrelativepath}:")
        flatten_child_histories(child_history, session, roothistorypath, childrelativepath)


@click.command()
@click.argument("root_path", required=False, type=click.Path(exists=True))
# options
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=True,
    type=click.Path(exists=True),
    help="Info for single file",
)
def info(verbose, single_file, root_path):
    """
    Prints information from the ASC MHL history

    \b
    """
    if single_file is not None and len(single_file) > 0:
        if root_path == None:
            current_dir = os.path.dirname(os.path.abspath(single_file[0]))
            while os.path.isdir(current_dir):
                asc_mhl_folder_path = os.path.join(current_dir, ascmhl_folder_name)
                if os.path.exists(asc_mhl_folder_path):
                    root_path = current_dir
                    break
                parent_dir = os.path.dirname(current_dir)
                # in case we get the same path again, we seem to be at the root
                # this works both on windows and unix
                if parent_dir == current_dir:
                    break
                current_dir = parent_dir

        if root_path is None:
            raise errors.NoMHLHistoryException(single_file[0])
        else:
            info_for_single_file(root_path, verbose, single_file)
        return
    else:
        info_for_entire_history(root_path, verbose)
    return


def info_for_entire_history(root_path, verbose):
    """
    ROOT_PATH: the root path to use for the asc mhl history
    """

    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.info(f"Info with history at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    log_child_histories(existing_history)


def log_child_histories(history):
    for hash_list in history.hash_lists:
        if logger.verbose_logging == True:
            creatorInfo = hash_list.creator_info.summary()
            processInfo = hash_list.process_info.summary()
            logger.info(
                f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})\n"
                f"     CreatorInfo: {creatorInfo}\n"
                f"     ProcessInfo: {processInfo}"
            )
        else:
            logger.info(f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})")

    for child_history in history.child_histories:
        logger.info(f"\nChild History at {child_history.get_root_path()}:")
        log_child_histories(child_history)


def info_for_single_file(root_path, verbose, single_file):
    """
    ROOT_PATH: the root path to use for the asc mhl history (optional)
    """

    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.info(f"Info with history at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    for path in single_file:
        relative_path = existing_history.get_relative_file_path(os.path.abspath(path))
        logger.info(f"{relative_path}:")
        for hash_list in existing_history.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            for hash_entry in media_hash.hash_entries:
                if logger.verbose_logging == True:
                    absolutePath = os.path.join(hash_list.get_root_path(), media_hash.path)
                    creatorInfo = hash_list.creator_info.summary()
                    processInfo = hash_list.process_info.summary()
                    logger.info(
                        f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})"
                        f" {hash_entry.hash_format}: {hash_entry.hash_string} ({hash_entry.action}) \n"
                        f"    {absolutePath}\n"
                        f"     CreatorInfo: {creatorInfo}\n"
                        f"     ProcessInfo: {processInfo}"
                    )
                    if media_hash.previous_path and relative_path == media_hash.path:
                        logger.info(
                            " In previous generations the file was named: {}\n\n".format(media_hash.previous_path)
                        )
                        info_for_single_file(root_path, verbose, [os.path.join(root_path, media_hash.previous_path)])
                else:
                    logger.info(
                        f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})"
                        f" {hash_entry.hash_format}: {hash_entry.hash_string} ({hash_entry.action})"
                    )


@click.command()
@click.argument("file_path", type=click.Path(exists=True))
# subcommands
@click.option(
    "--directory_file",
    "-df",
    default=False,
    is_flag=True,
    help="Check directory file (e.g. ascmhl_chain.xml) instead of manifest file",
)
@click.option(
    "--xsd_file",
    "-xsd",
    default=None,
    help="Path to the xsd file",
)
def xsd_schema_check(file_path, directory_file, xsd_file):
    """
    Checks a .mhl file against the xsd schema definition

    \b
    The xsd-schema-check command validates a given ASC MHL file against the XML
    XSD. This command can be used to ensure the creation of syntactically valid
    ASC MHL files, for example during implementation of tools creating ASC MHL
    files.
    """

    xsd_path = "xsd/ASCMHL.xsd"

    if directory_file:
        xsd_path = "xsd/ASCMHLDirectory__combined.xsd"

    if xsd_file != None:
        xsd_path = xsd_file

    xsd = etree.XMLSchema(etree.parse(xsd_path))

    # pass a file handle to support the fake file system used in the tests
    file = open(file_path, "rb")
    result = xsd.validate(etree.parse(file))
    file.close()

    if result:
        logger.info(f"validated: {file_path}")
    else:
        logger.error(f"ERROR: {file_path} didn't validate against XSD!")
        logger.info(f"Issues:\n{xsd.error_log}")
        raise errors.VerificationFailedException


def test_for_missing_files(not_found_paths, root_path, ignore_spec: MHLIgnoreSpec = MHLIgnoreSpec()):
    ignore_path_spec = ignore_spec.get_path_spec()
    # update to exclude our ignored files
    not_found_paths = [x for x in not_found_paths if not ignore_path_spec.match_file(x)]
    if len(not_found_paths) == 0:
        return None
    # test our not_found_paths against our ignore spec to ensure these weren't explicitly ignored.
    logger.error(f"ERROR: {len(not_found_paths)} missing file(s):")
    for path in not_found_paths:
        logger.error(f"  {os.path.relpath(path, root_path)}")
    return errors.CompletenessCheckFailedException()


def commit_session(session, author_name, author_email, author_phone, author_role, location, comment):
    creator_info = MHLCreatorInfo()
    creator_info.tool = MHLTool(ascmhl_tool_name, ascmhl_tool_version)
    creator_info.creation_date = utils.datetime_now_isostring()
    creator_info.host_name = platform.node()
    creator_info.location = location
    creator_info.comment = comment
    if author_name is not None or author_email is not None or author_role is not None or author_phone is not None:
        author_object = MHLAuthor(author_name, author_email, author_phone, author_role)
        creator_info.authors.append(author_object)

    process_info = MHLProcessInfo()
    process_info.process = MHLProcess("in-place")

    session.commit(creator_info, process_info)


def commit_session_for_collection(
    session, root_path, author_name, author_email, author_phone, author_role, location, comment
):
    creator_info = MHLCreatorInfo()
    creator_info.tool = MHLTool(ascmhl_tool_name, ascmhl_tool_version)
    creator_info.creation_date = utils.datetime_now_isostring()
    creator_info.host_name = platform.node()
    creator_info.location = location
    creator_info.comment = comment
    if author_name is not None:
        author_object = MHLAuthor(author_name, author_email, author_phone, author_role)
        creator_info.authors.append(author_object)

    process_info = MHLProcessInfo()
    process_info.process = MHLProcess("flatten")
    root_hash = MHLMediaHash()
    root_hash.path = root_path
    process_info.root_media_hash = root_hash
    process_info.hashlist_custom_basename = "packinglist_" + os.path.basename(root_path)

    session.commit(creator_info, process_info, writeChain=False)


"""
A tuple for returning the result of a seal file path operation
attributes:
hash_value -- string value, a hash
success -- boolean value, indicates if the update was successful
"""
SealPathResult = namedtuple("SealPathResult", ["hash_value", "success"])


def seal_file_path(existing_history, file_path, hash_formats: [str], session) -> Dict[str, SealPathResult]:
    """
    Generates hashes for a file path.
    Compares the generated hashes to any existing hash records
    Adds the generated hashes to the hash history of the file path
    :param existing_history: The existing hash record
    :param file_path: The path for which to generate hashes
    :param hash_formats: The hash formats to generate
    :param session: The session to which the generated hashes will be added
    :return: A dictionary keyed by hash_format strings.
    Each entry contains the hash value and a boolean indicating if updating was successful
    """
    relative_path = existing_history.get_relative_file_path(file_path)
    file_size = os.path.getsize(file_path)
    file_modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(file_path))

    # find in the according child history the already available hash formats
    existing_child_history, existing_history_relative_path = existing_history.find_history_for_path(relative_path)
    existing_hash_formats = existing_child_history.find_existing_hash_formats_for_path(existing_history_relative_path)

    # create a separate list of hash formats for which hashes will be generated
    hash_formats_to_generate = []

    # if there are existing entries for this path, the recorded formats must also be generated
    if existing_hash_formats and len(existing_hash_formats) > 0:
        # existing hash formats should be verified prior to any new formats
        for hash_format in existing_hash_formats:
            if hash_format in hash_formats:
                hash_formats_to_generate.append(hash_format)
        # if no formats are being carried over from the previous generation to this one, at least
        # one of the previous generation hashes needs to at least be verified as correct
        # TODO: Consider making the selection of the previous generation bench mark format less arbitrary
        # TODO: Instead of selecting the first format, perhaps choose the most efficient or robust format
        if not hash_formats_to_generate or len(hash_formats_to_generate) == 0:
            hash_formats_to_generate.append(existing_hash_formats[0])

    for hash_format in hash_formats:
        if hash_format not in hash_formats_to_generate:
            hash_formats_to_generate.append(hash_format)

    # generate the file hashes
    current_hash_lookup = multiple_format_hash_file(file_path, hash_formats_to_generate)

    # the lookup where the results will be stored
    hash_result_lookup = {}

    # any newly generated hash will require any previous hashes to be verified
    existing_hashes_verified = True

    # verify the existing hash entries first
    if existing_hash_formats:
        for hash_format in existing_hash_formats:
            # make sure the existing format was included in the hashes which were generated
            if hash_format not in hash_formats_to_generate:
                continue

            success = True
            success &= session.append_file_hash(
                file_path, file_size, file_modification_date, hash_format, current_hash_lookup[hash_format]
            )

            # if even one existing hash failed to verify existing hashes will be deemed corrupt
            if not success:
                existing_hashes_verified = False

            # if the existing format was requested add it to the results lookup
            if hash_format in hash_formats:
                hash_result_lookup[hash_format] = SealPathResult(current_hash_lookup[hash_format], success)

    # add the new hash formats
    for hash_format in hash_formats_to_generate:
        # existing hashes have already been handled
        if existing_hash_formats and hash_format in existing_hash_formats:
            continue

        # if this format has never been recorded the previous hashes must be verified for this hash to be verified
        success = existing_hashes_verified

        # only add the new hash format to the session if the previous hashes are verified
        if success:
            success &= session.append_file_hash(
                file_path, file_size, file_modification_date, hash_format, current_hash_lookup[hash_format]
            )

        # If the existing hash was requested by the caller it needs to be added to the lookup
        if hash_format in hash_formats:
            hash_result_lookup[hash_format] = SealPathResult(current_hash_lookup[hash_format], success)

    return hash_result_lookup


def get_ignore_spec_including_nested_ignores(existing_history, ignore_list, ignore_spec_file=None):
    """Get the ignore patterns from nested histories with their respective paths,
    so that ignored files from nested histories are also ignored in this session, but are not stored
    in the root ascmhl manifest"""
    ignore_patterns_cumulated = ignore.default_ignore_list()
    # handle non-existent ignores in root history
    if existing_history.latest_ignore_patterns() is not None:
        for x in existing_history.latest_ignore_patterns():
            if x not in ignore.default_ignore_list():
                ignore_patterns_cumulated.append(x)

    for x in existing_history.latest_ignore_pattern_from_nested_histories():
        ignore_patterns_cumulated.append(x)

    for x in ignore_list:
        ignore_patterns_cumulated.append(x)

    patterns_from_file = []
    if ignore_spec_file:
        with open(ignore_spec_file, "r") as fh:
            patterns_from_file.extend(line.rstrip("\n") for line in fh if line != "\n")
        for x in patterns_from_file:
            if x not in ignore_patterns_cumulated:
                ignore_patterns_cumulated.append(x)

    # we now build the absolute ignore paths for the current session from all nested ignores
    # otherwise the post_order_lexicographic() won't ignore these paths
    absolute_ignore_paths = []
    path = existing_history.get_root_path()
    for pattern in ignore_patterns_cumulated:
        if pattern in ignore.default_ignore_list():
            absolute_ignore_paths.append(pattern)
        else:
            if pattern.find("/") != -1:
                if pattern.startswith("/"):
                    absolute_ignore_paths.append(path + pattern)
                elif pattern.startswith("**"):
                    absolute_ignore_paths.append(pattern)
                elif pattern.endswith("/") and pattern[:-1].find("/") == -1:
                    absolute_ignore_paths.append(pattern)
                elif pattern.endswith("/" + "**") and pattern[:-3].find("/") == -1:
                    absolute_ignore_paths.append(pattern)
                else:
                    absolute_ignore_paths.append(path + os.sep + pattern)
            else:
                absolute_ignore_paths.append(pattern)

    normalized_paths = [pathspec.util.normalize_file(p) for p in absolute_ignore_paths]
    spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), normalized_paths)
    return spec



from packaging import version

from ascmhl.utils import datetime_now_filename_string, datetime_now_isostring
from freezegun import freeze_time
import datetime
import os, time


def test_time():
    datetime_now_filename = datetime_now_filename_string()
    datetime_now_filename_reference = datetime.datetime.strftime(
        datetime.datetime.now(datetime.timezone.utc), "%Y-%m-%d_%H%M%SZ"
    )

    print("datetime_now_filename:           " + datetime_now_filename)
    print("datetime_now_filename_reference: " + datetime_now_filename_reference)

    # seems always to be working, as tests run in UTC ?
    assert datetime_now_filename == datetime_now_filename_reference


@freeze_time("2020-01-15 13:00:00")
def test_freeze_time():
    iso_string = datetime_now_isostring()
    assert iso_string == "2020-01-15T13:00:00+00:00"



"""
__author__ = "Jon Waggoner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Jon Waggoner"
__email__ = "opensource@jonwaggoner.com"
"""

from os.path import join, isdir
import os

from . import logger
import pathspec
from .__version__ import ascmhl_folder_name


def post_order_lexicographic(top: str, ignore_pathspec: pathspec.PathSpec = None):
    """
    iterates a file system in the order necessary to generate composite tree hashes, bypassing ignored paths.

    :param top: the directory being iterated
    :param ignore_pathspec: the pathspec of ignore patterns to match file exclusions against
    :return: yields results in folder chunks, in the order necessary for composite directory hashes
    """
    # create a sorted list of our immediate children
    names = os.listdir(top)
    names.sort()

    # list of tuples. each tuple contains the child name and whether the child is a directory.
    children = []
    for name in names:
        file_path = os.path.join(top, name)
        is_directory = isdir(file_path)
        if is_directory:
            file_path = file_path + "/"
        if ignore_pathspec and ignore_pathspec.match_file(file_path):
            if os.path.basename(os.path.normpath(file_path)) != ascmhl_folder_name:
                logger.verbose(f"ignoring filepath {file_path.rstrip('/')}")
            continue
        children.append((name, is_directory))

    # if directory, yield children recursively in post order until exhausted.
    for name, is_dir in children:
        if is_dir:
            path = join(top, name)
            if not os.path.islink(path):
                for x in post_order_lexicographic(path, ignore_pathspec):
                    yield x

    # now that all children have been traversed, yield the top (current) directory and all of it's sorted children.
    yield top, children



"""
__author__ = "Jon Waggoner, Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner"
__email__ = "opensource@pomfort.com"
"""

import sys
import click

verbose_logging = False
debug_logging = False


def debug(msg, *args):
    """Logs a message to stdout only if debug is enabled."""
    if debug_logging:
        info(msg, *args)


def verbose(msg, *args):
    """Logs a message to stdout only if verbose is enabled."""
    if verbose_logging:
        info(msg, *args)


def info(msg, *args):
    """Logs a message to stdout."""
    if args:
        msg %= args
    click.echo(msg, file=sys.stdout)


def error(msg, *args):
    """Logs a message to stderr"""
    if args:
        msg %= args
    click.echo(click.style(msg, fg="red", bold=True), file=sys.stderr)


def fatal(msg, *args):
    """Logs a message to stderr, then exits"""
    if args:
        msg %= args
    click.echo(click.style(msg, fg="red", bold=True, blink=True), file=sys.stderr)
    click.get_current_context().abort()



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from .conftest import abspath_conversion_tests
from .conftest import path_conversion_tests

from freezegun import freeze_time
from click.testing import CliRunner

from ascmhl.history import MHLHistory
import ascmhl.commands

scenario_output_path = "examples/scenarios/Output"
fake_ref_path = "/ref"


@freeze_time("2020-01-16 09:15:00")
def test_record_succeed_single_file(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-sf", abspath_conversion_tests("/root/A/A1.txt")]
    )
    assert result.exit_code == 0
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")

    # make sure that only the specified file was added
    history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    assert len(history.hash_lists[0].media_hashes) == 1
    assert history.hash_lists[0].media_hashes[0].path == str(path_conversion_tests("A/A1.txt"))


@freeze_time("2020-01-16 09:15:00")
def test_record_succeed_single_directory(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-sf", abspath_conversion_tests("/root/A")]
    )
    assert result.exit_code == 0
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")

    # make sure that only the specified file was added
    history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    assert len(history.hash_lists[0].media_hashes) == 2
    assert history.hash_lists[0].media_hashes[0].path == str(path_conversion_tests("A/A1.txt"))
    assert history.hash_lists[0].media_hashes[1].path == str(path_conversion_tests("A/A2.txt"))


@freeze_time("2020-01-16 09:15:00")
def test_record_succeed_multiple_files(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [
            abspath_conversion_tests("/root"),
            "-sf",
            abspath_conversion_tests("/root/A/A1.txt"),
            "-sf",
            abspath_conversion_tests("/root/A/A2.txt"),
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")

    # make sure that only the specified file was added
    history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    assert len(history.hash_lists[0].media_hashes) == 2
    assert history.hash_lists[0].media_hashes[0].path == str(path_conversion_tests("A/A1.txt"))
    assert history.hash_lists[0].media_hashes[1].path == str(path_conversion_tests("A/A2.txt"))


def test_record_fail_altered_file(fs, simple_mhl_history):
    # alter a file
    with open("/root/Stuff.txt", "a") as file:
        file.write("!!")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-sf", abspath_conversion_tests("/root/Stuff.txt")]
    )
    assert result.exit_code == 11
    assert "Stuff.txt" in result.output

    # when passing a different file to record no error ws thrown since the altered file is ignored
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-sf", abspath_conversion_tests("/root/A/A1.txt")]
    )
    assert result.exit_code == 0

    # make sure we have created one failing and one succeeded generation
    history = MHLHistory.load_from_path(abspath_conversion_tests("/root"))
    assert history.hash_lists[1].media_hashes[0].path == "Stuff.txt"
    assert history.hash_lists[1].media_hashes[0].hash_entries[0].action == "failed"
    assert history.hash_lists[2].media_hashes[0].path == str(path_conversion_tests("A/A1.txt"))
    assert history.hash_lists[2].media_hashes[0].hash_entries[0].action == "verified"



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from __future__ import annotations
import os
import re
from datetime import datetime, date, time

from . import hasher, ignore
from .__version__ import ascmhl_folder_name, ascmhl_file_extension, ascmhl_chainfile_name, ascmhl_collectionfile_name
from . import hashlist_xml_parser, chain_xml_parser
from .utils import datetime_now_filename_string
from typing import Tuple, List, Dict, Optional, Set
from . import logger, errors
from .chain import MHLChain
from .hashlist import MHLHashList, MHLHashEntry


class MHLHistory:
    """
    class for representing an entire MHL history

    - public interface
        * intitialized with a root path containting an MHL folder with MHL files
        * accessing derived information (e.g. earliest hash for file)

    - private interface
        * initialize new, empty MHLHashList for adding files/hashes, define path and filename for new MHLHashList

    model member variables:
    chain -- content of the chain file
    hashlists -- list of MHLHashList (one for each generation / MHL file)
    child_histories -- list of all direct child histories
    parent_history -- the one parent history if any
    child_history_mappings -- mapping of all (also transitive) child histories and their relative
                              path to find the appropriate child

    attribute member variables:

    other member variables:
    root_path -- path where the mhl folder resides
    """

    history_file_name_regex = r"^(\d{4,})(?:_(.+))?$"

    chain: Optional[MHLChain]
    hash_lists: List[MHLHashList]
    asc_mhl_path: Optional[str]
    child_histories: List[MHLHistory]
    child_history_mappings: Dict[str, MHLHistory]
    parent_history: Optional[MHLHistory]

    def __init__(self):
        self.chain = None
        self.hash_lists = []
        self.child_histories = []
        self.child_history_mappings = {}
        self.parent_history = None
        self.asc_mhl_path = None

    def append_hash_list(self, hash_list):
        self.hash_lists.append(hash_list)

    def get_root_path(self):
        if not self.asc_mhl_path:
            return None
        return os.path.dirname(self.asc_mhl_path)

    def get_relative_file_path(self, file_path):
        if not self.asc_mhl_path:
            return None
        if os.path.isabs(file_path):
            return os.path.relpath(file_path, os.path.dirname(self.asc_mhl_path))
        return None

    def latest_generation_number(self) -> int:
        latest_number = 0
        for hash_list in self.hash_lists:
            if hash_list.generation_number:
                latest_number = hash_list.generation_number
        return latest_number

    def latest_ignore_patterns(self) -> Optional[List[str]]:
        """
        latest_ignore_patterns will return the list of ignore patterns from the latest generation.
        returns None if does not exist.
        """
        if len(self.hash_lists) == 0:
            return None
        hash_list = self.hash_lists[-1]
        if not hash_list or not hash_list.process_info.ignore_spec:
            return None
        return hash_list.process_info.ignore_spec.get_pattern_list()

    def latest_ignore_pattern_from_nested_histories(self) -> Optional[List[str]]:
        parent_path = self.get_root_path()
        cumulated_ignores = []
        for path, history in self.child_history_mappings.items():
            for pattern in history.latest_ignore_patterns():
                # don't add the default pattern
                child_path = history.get_root_path()
                path = os.path.relpath(child_path, parent_path)
                if pattern in ignore.default_ignore_list():
                    continue
                else:
                    # return the directory of the history with the pattern appended
                    if pattern.find("/") != -1:
                        cumulated_ignores.append(path + pattern)
                    else:
                        cumulated_ignores.append(path + "/**/" + pattern)

        return cumulated_ignores

    # methods to query and compare hashes
    def find_original_hash_entry_for_path(self, relative_path: str) -> Optional[MHLHashEntry]:
        """Searches the history for the first (original) hash of a file

        starts with the first generation, if we don't find it there we continue to look in all other generations
        until we've found the first appearance of the give file.
        """
        for hash_list in self.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            for hash_entry in media_hash.hash_entries:
                if hash_entry.action == "original":
                    return hash_entry
        return None

    # methods to query and compare hashes
    def find_directory_hash_entries_for_path(self, relative_path: str) -> List[MHLHashEntry]:
        """Searches the history for directory hash entries of a folder

        starts with the first generation through all other generations
        and collects all directory hashes found for the given folder.
        """
        directory_hash_entries = []
        for hash_list in self.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            if media_hash.is_directory:
                for hash_entry in media_hash.hash_entries:
                    # FIXME is there a better way of accessing the generation from a hash entry?
                    hash_entry.temp_generation_number = hash_list.generation_number
                directory_hash_entries = directory_hash_entries + media_hash.hash_entries

        # also search the root directory hashes from all child histories
        if relative_path == ".":
            for hash_list in self.hash_lists:
                for hash_entry in hash_list.process_info.root_media_hash.hash_entries:
                    # FIXME is there a better way of accessing the generation from a hash entry?
                    hash_entry.temp_generation_number = hash_list.generation_number
                    hash_entry.temp_is_root_folder = True
                directory_hash_entries = directory_hash_entries + hash_list.process_info.root_media_hash.hash_entries

        return directory_hash_entries

    def find_first_hash_entry_for_path(self, relative_path, hash_format=None) -> Optional[MHLHashEntry]:
        """Searches the history for the first (original) hash entry of a file
        or if an optional hash format is given the first hash in that format

        starts with the first generation, if we don't find it there we continue to look in all other generations
        until we've found the first appearance of the give file.
        """
        for hash_list in self.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            for hash_entry in media_hash.hash_entries:
                if hash_format is not None and hash_entry.hash_format == hash_format:
                    return hash_entry
                elif hash_format is None:
                    return hash_entry
        return None

    def find_existing_hash_formats_for_path(self, relative_path: str) -> List[str]:
        """Searches through the history to find all existing hash formats we might want to compare against"""
        hash_formats = []
        for hash_list in self.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            for hash_entry in media_hash.hash_entries:
                if hash_entry.hash_format not in hash_formats:
                    hash_formats.append(hash_entry.hash_format)
        return hash_formats

    # def handling of child histories
    def find_history_for_path(self, relative_path: str) -> Tuple[MHLHistory, str]:
        if len(self.child_histories) == 0:
            return self, relative_path
        dir_path = relative_path
        # shorten the path until we find a mapping otherwise there is no child history that should handle the path
        while len(dir_path) > 0:
            if dir_path in self.child_history_mappings:
                history = self.child_history_mappings[dir_path]
                absolute_path = os.path.join(self.get_root_path(), relative_path)
                history_relative_path = history.get_relative_file_path(absolute_path)
                return history, history_relative_path
            dir_path = os.path.dirname(dir_path)
        return self, relative_path

    def set_of_file_paths(self) -> Set[str]:
        all_paths = set()
        for hash_list in self.hash_lists:
            all_paths.update(hash_list.set_of_file_paths(self.get_root_path()))
        for child_history in self.child_histories:
            all_paths.update(child_history.set_of_file_paths())

        return all_paths

    def renamed_path_with_previous_path(self):
        all_paths = {}
        for hash_list in self.hash_lists:
            all_paths.update(hash_list.renamed_path_with_previous_path(self.get_root_path()))
        for child_history in self.child_histories:
            all_paths.update(child_history.renamed_path_with_previous_path())
        return all_paths

    def hash_list_with_file_name(self, file_name) -> Optional[MHLHashList]:
        for hash_list in self.hash_lists:
            if hash_list.get_file_name() == file_name:
                return hash_list
        return None

    def append_child_history(self, child_history: MHLHistory) -> None:
        self.child_histories.append(child_history)

    # loading history and child histories from path

    @classmethod
    def load_from_path(cls, root_path):
        """finds all MHL files in the asc-mhl folder, returns the mhl_history instance with all mhl_hashlists"""

        asc_mhl_folder_path = os.path.join(root_path, ascmhl_folder_name)
        history = cls()
        history.asc_mhl_path = asc_mhl_folder_path

        file_path = os.path.join(asc_mhl_folder_path, ascmhl_chainfile_name)
        if os.path.exists(asc_mhl_folder_path) and not os.path.exists(file_path):
            raise errors.NoMHLChainException(file_path)
        history.chain = chain_xml_parser.parse(file_path)
        if history.chain.generations:
            for generation in history.chain.generations:
                expected_file = os.path.join(asc_mhl_folder_path, generation.ascmhl_filename)
                if os.path.exists(expected_file):
                    hash = hasher.hash_file(expected_file, generation.hash_format)
                    if hash != generation.hash_string:
                        raise errors.ModifiedMHLManifestFileException(expected_file)
                else:
                    raise errors.MissingMHLManifestException(expected_file)

        hash_lists = []
        for root, directories, filenames in os.walk(asc_mhl_folder_path):
            for filename in filenames:
                # file name example: 0001_root_2020-01-15_130000.mhl
                # ignore ._ variants of mhl files that can happen when moving data from macOS to Windows and back
                if (len(filename) > 2 and filename[:2] == "._") or not filename.endswith(ascmhl_file_extension):
                    continue
                filename_no_extension, _ = os.path.splitext(filename)
                parts = re.findall(MHLHistory.history_file_name_regex, filename_no_extension)
                if len(parts) == 1 and len(parts[0]) == 2:
                    file_path = os.path.join(asc_mhl_folder_path, filename)
                    hash_list = hashlist_xml_parser.parse(file_path)
                    generation_number = int(parts[0][0])
                    hash_list.generation_number = generation_number
                    # FIXME is there a better way of accessing the generation from a hash entry?
                    if hash_list.process_info.root_media_hash is not None:
                        for hash_entry in hash_list.process_info.root_media_hash.hash_entries:
                            hash_entry.temp_generation_number = hash_list.generation_number
                    hash_lists.append(hash_list)
                else:
                    logger.error(f"name of ascmhl file {filename} does not conform to naming convention")
        # sort all found hash lists by generation number first to make sure we add them to the history in order
        hash_lists.sort(key=lambda x: x.generation_number)
        for hash_list in hash_lists:
            history.append_hash_list(hash_list)

        history._find_and_load_child_histories()

        return history

    @classmethod
    def load_from_packing_list_path(cls, packing_list_path, root_path):
        """returns the mhl_history instance with the one packing list mhl_hashlists"""
        # via https://docs.google.com/document/d/1FVSyHq2XJdNt-3Vur_5I_FPOoeeC_cEjkv7p---biyg/edit#

        asc_mhl_folder_path = os.path.join(root_path, ascmhl_folder_name)
        history = cls()
        history.asc_mhl_path = asc_mhl_folder_path

        hash_list = hashlist_xml_parser.parse(packing_list_path)
        hash_list.generation_number = 1
        history.append_hash_list(hash_list)

        return history

    @classmethod
    def create_collection_at_path(cls, root_path, debug=False):
        now = datetime.now()
        collection_folder_name = "debug"
        if not debug:
            date_time_string = now.strftime("%Y-%m-%d")
            collection_folder_name = "collection_" + date_time_string

        collection_folder_path = os.path.join(root_path, collection_folder_name)

        parent_path = os.path.dirname(collection_folder_path)
        if not os.path.isdir(parent_path):
            os.mkdir(parent_path)
        if not os.path.isdir(collection_folder_path):
            os.mkdir(collection_folder_path)

        history = cls()
        history.asc_mhl_path = collection_folder_path

        file_path = os.path.join(collection_folder_path, ascmhl_collectionfile_name)
        history.chain = chain_xml_parser.parse(file_path)

        return history

    def _find_and_load_child_histories(self) -> None:
        """traverses the whole file system tree inside the history to find all sub histories"""
        history_root = self.get_root_path()
        for root, directories, _ in os.walk(history_root):
            if root != history_root and ascmhl_folder_name in directories:
                # we parse the mhl folder and clear the directories so we are not going deeper
                # everything beneath is handled by the child history
                child_history = MHLHistory.load_from_path(root)
                child_history.parent_history = self
                self.append_child_history(child_history)
                directories.clear()

        # update parent children mapping with the found children
        self._update_child_history_mapping()
        self._resolve_hash_list_references()

    def _update_child_history_mapping(self) -> None:
        self.child_history_mappings = {}
        for child_history in self.child_histories:
            relative_child_path = self.get_relative_file_path(child_history.get_root_path())
            self.child_history_mappings[relative_child_path] = child_history
            for sub_child_relative_path, sub_child in child_history.child_history_mappings.items():
                relative_path = os.path.join(relative_child_path, sub_child_relative_path)
                self.child_history_mappings[relative_path] = sub_child

        if self.parent_history is not None:
            self.parent_history._update_child_history_mapping()

    @staticmethod
    def walk_child_histories(history: MHLHistory):
        for child in history.child_histories:
            yield from MHLHistory.walk_child_histories(child)
        yield history

    def _resolve_hash_list_references(self) -> None:
        """for all hash lists resolve existing hash list references by searching them in the child histories"""
        for hash_list in self.hash_lists:
            for reference in hash_list.hash_list_references:
                reference_path = os.path.dirname(os.path.dirname(reference.path))
                history = self.child_history_mappings.get(reference_path, None)
                if history is None:
                    return
                referenced_hash_list = history.hash_list_with_file_name(os.path.basename(reference.path))
                assert referenced_hash_list is not None
                generated_hash = referenced_hash_list.generate_reference_hash()
                if generated_hash != reference.reference_hash:
                    logger.error("ASC MHL file: " + hash_list.file_path)
                    logger.error('  referenced manifest file "' + referenced_hash_list.file_path + '"')
                    logger.error("  should have hash:  " + reference.reference_hash)
                    logger.error("  but file has hash: " + generated_hash)

                hash_list.referenced_hash_lists.append(referenced_hash_list)

    # writing new generations

    def write_new_generation(self, new_hash_list: MHLHashList):
        self._validate_new_hash_list(new_hash_list)
        file_name, generation_number = self._new_generation_filename()
        if new_hash_list.process_info.hashlist_custom_basename is not None:
            file_name = self._new_custom_filename(new_hash_list.process_info.hashlist_custom_basename)
        file_path = os.path.join(self.asc_mhl_path, file_name)
        new_hash_list.generation_number = generation_number
        hashlist_xml_parser.write_hash_list(new_hash_list, file_path)
        self.append_hash_list(new_hash_list)

    def _new_custom_filename(self, custom_basename):
        date_string = datetime_now_filename_string()
        file_name = f"{custom_basename}_{date_string}{ascmhl_file_extension}"
        return file_name

    def _new_generation_filename(self):
        date_string = datetime_now_filename_string()
        index = self.latest_generation_number() + 1
        folder_name = os.path.basename(os.path.normpath(self.get_root_path()))
        file_name = f"{index:04d}_{folder_name}_{date_string}{ascmhl_file_extension}"
        return file_name, index

    def _validate_new_hash_list(self, hash_list):
        """Method to check mandatory conditions e.g. before serializing
        Validates for example that a new hash in a different format is verified
        against the original hash(format) of the same file
        """

        # TODO: validate new hash entries
        for media_hash in hash_list.media_hashes:
            for hash_entry in media_hash.hash_entries:
                if hash_entry.action == "new":
                    # TODO: do need to use the original hash here or can we also use another hash
                    original_hash_entry = self.find_original_hash_entry_for_path(media_hash.path)
                    required_hash_entry = media_hash.find_hash_entry_for_format(original_hash_entry.hash_format)
                    if required_hash_entry is None:
                        raise AssertionError("no hash entry found for new hash", hash_entry)
                    if required_hash_entry.action != "verified":
                        raise AssertionError("hash entry for new hash not verified", hash_entry, required_hash_entry)
                    hash_entry.action = "verified"
        return True

    # accessors
    def log(self):
        logger.info("mhl history")
        logger.info("asc_mhl path: {0}".format(self.asc_mhl_path))

        for hash_list in self.hash_lists:
            logger.info("")
            hash_list.log()
